/************************************************************************
  The Blade Libraries (c) Blade 2001/2.
  PSX programming libraries written in C.

*************************************************************************/

#include <psone/gpu.h>
#include <psone/specfont.h>
#include <psone/stdarg.h>
#include <psone/math.h>
#include <psone/syscall.h>

//=======================================================
typedef volatile long * vptr;

#define	i_reg	(*(vptr)0x1f801070)
#define	i_mask	(*(vptr)0x1f801074)
#define	d2_madr	(*(vptr)0x1f8010a0)
#define	d2_bcr	(*(vptr)0x1f8010a4)
#define	d2_chcr	(*(vptr)0x1f8010a8)
#define	d_pcr	(*(vptr)0x1f8010f0)
#define	d_icr	(*(vptr)0x1f8010f4)
#define	gp0	(*(vptr)0x1f801810)
#define	gp1	(*(vptr)0x1f801814)
#define d2_linear (0x01000201)
#define d2_linked (0x01000401)

#define wait_gpu while (!(gp1 & 0x04000000))
#define wait_dma while (d2_chcr & 0x01000000)
//=======================================================
void GP_SendGPU0(u_long gpu_primitive);
void Mem2Vram(TIMDATA *pTIM, u_long size_in_bytes);
int  PutNum(char *out, int n, int w, int padc, int radix);
void PrintChar(u_char num, int x, int y, u_char ascii);
void GsSortSprite2(GsSPRITE *sprite);
void object_quick_sort(int, int);
//=======================================================
u_long VIDEOMODE = 0;			// video mode: (0:NTSC, 1:PAL)
u_long PSDCNT = 0;			// frame switch counter
u_long PSDOFSX[2];			// double buffer x coords
u_long PSDOFSY[2];			// double buffer y coords
u_long GP_LENGTH = 0;			// length of primitive queue
u_long * G_GP = NULL;			// pointer to primitive queue
u_long FONTCOLOR[10];			// font colors
u_char STRINGBUFF[128];			// buffer for FntPrint
GsIMAGE FONT[10];			// font information
u_char FONT_WIDTH[10];			// font width
u_char FONT_HEIGHT[10];			// font height
//=======================================================
u_long OBJECT_number = 0;
u_long OBJECT_z[1024];
u_long * OBJECT_prev;
u_long * OBJECT_first;
u_long * OBJECT_start[1024];
u_long * OBJECT_end[1024];
//=======================================================
// macros for handling packets
//=======================================================
static inline void ADD_PACKET (int num, u_long data)
{
  *(G_GP + num + 1) = data;
}
//-------------------------------------------------------
static inline void NEXT_PACKET (int size)
{
  u_long data;
  u_long address = (u_long)(G_GP + size + 1) & 0x00ffffff;
  
  data = (size << 24) | address;

  *(G_GP) = data;

  OBJECT_prev = G_GP; 

  G_GP += (size + 1);  

	//  printf("%08x > (G_GP + 0) = %08x\n", (G_GP), data );  
	//  for (i = 0; i < size; i++) {
    	//printf("%08x - (G_GP + %d) = %08x\n", (G_GP + i + 1), i + 1, *(G_GP+i+1) );
    	//}
}
//-------------------------------------------------------
static inline void SWITCH_ADDRESS (u_long * first, u_long * second)
{
  *first = (*first & 0xff000000) | ((u_long)second & 0x00ffffff);
}
//=======================================================
// void GsInitGraph(int x_res, int y_res, int intl, int dither, int vram);
// x-res = horizontal resolution (256/320/384/512/640)
// y-res = vertical (240/480 NTSC, 256,512 PAL)
// intl = bit 0 - interlace flag (0:off, 1:on)
// dither = dither flag (0:off, 1:on)
// vram = frame buffer mode (0:16bit, 1:24bit)
//-------------------------------------------------------
void GsInitGraph(int x_res, int y_res, int intl, int dither, int vram)
{	
  /* reset the GPU */
  gp1 = 0x00000000;

  /* enable linked-list DMA transfer */
  gp1 = 0x04000002;

  /* enable DMA channel 2 */
  d_pcr |= 0x0800;


  /* output a message to the PC console */
  printf("Blade Libraries v1.2\n");

  /* initialise some values */
  GsDISPENV.disp.x = 0;
  GsDISPENV.disp.y = 0;
  GsDISPENV.disp.w = x_res;
  GsDISPENV.disp.h = y_res;

  GsDRAWENV.tw.x = 0;
  GsDRAWENV.tw.y = 0;
  GsDRAWENV.tw.w = 0;
  GsDRAWENV.tw.h = 0;

  GsDRAWENV.clip.x = 0;
  GsDRAWENV.clip.y = 0;
  GsDRAWENV.clip.w = 1023;
  GsDRAWENV.clip.h = 511;

  GsDISPENV.screen.x = 0;
  GsDISPENV.screen.y = 0;
  GsDISPENV.screen.w = 0;
  GsDISPENV.screen.h = 0;

  if (VIDEOMODE == MODE_PAL)
    GsDISPENV.screen.y = 23;

  GsDISPENV.isinter = (intl & 0x01) << 5;
  GsDISPENV.isrgb24 = (vram & 0x01) << 4;

  GsDRAWENV.ofs[0] = 0;
  GsDRAWENV.ofs[1] = 0;
  GsDRAWENV.tpage = 0;
  GsDRAWENV.dtd = (dither & 0x01) << 9;
  GsDRAWENV.dfe = 1 << 10;
  GsDRAWENV.isbg = 0;

  PutDrawEnv(&GsDRAWENV);
  PutDispEnv(&GsDISPENV);

  FntLoad(0, (TIMHEADER *) specfont);

  gp1 = 0x03000000;			// enable the display
}
//=======================================================
void GsInitGraphQuick(int x_res, int y_res)
{	
  SetVideoMode(MODE_AUTO);

  GsInitGraph(x_res, y_res, 0, 1, 0);

  GsDefDispBuff(0, 0, 0, y_res);
}
//=======================================================
DRAWENV *PutDrawEnv(DRAWENV *env)
{	
  GP_SendGPU0 (0xe1000200);				// drawing mode

  GP_SendGPU0 (0xe2000000);				// texture window settings

  GP_SendGPU0 (0xe3000000 | env->clip.y << 10 | env->clip.x);	// clip area start and end
  GP_SendGPU0 (0xe4000000 | (env->clip.h + env->clip.y) << 10 | (env->clip.w + env->clip.x));

  GP_SendGPU0 (0xe5000000);				// drawing offset env->ofs[0]  env->ofs[1]

  GP_SendGPU0 (0xe6000000);  				// mask settings

  return (env);
}
//=======================================================
DISPENV *PutDispEnv(DISPENV *env)
{	
  u_long x0, y0, x1, y1;

  // display offset
  gp1 = (0x05000000 | (env->disp.y << 10) | (env->disp.x & 0xffff));

  // horizontal screen start/end
  x0 = env->screen.x *10 + 608;
  x1 = ((env->screen.x + 256) *10) + 608;
  gp1 = (0x06000000 | x1 << 12 | x0);

  // vertical screen start/end
  y0 = env->screen.y + 16 + (3 *VIDEOMODE);
  y1 = y0 + 240;
  gp1 = (0x07000000 | y1 << 10 | y0);

  // display mode
  switch (env->disp.w) {
    case 256:	x0 = ENV_W256;
    		break;
    case 384:	x0 = ENV_W384;
    		break;
    case 512:	x0 = ENV_W512;
    		break;
    case 640:	x0 = ENV_W640;
    		break;
    default:	x0 = ENV_W320;
    }

  y0 = ENV_H240;
  if (env->disp.h == 480)
    y0 = ENV_H480;

  gp1 = (0x08000000 | x0 | y0 | env->isinter | env->isrgb24 | (VIDEOMODE << 3) );
  return (env);
}
//=======================================================
void GsDefDispBuff(int x0, int y0, int x1, int y1)
{	
  if (y0 > 255) 
    y0 = 255;

  if (y1 > 255)
    y1 = 255;

  PSDOFSX[0] = x0;
  PSDOFSY[0] = y0;
  PSDOFSX[1] = x1;
  PSDOFSY[1] = y1;
}
//=======================================================
int GsGetActiveBuff(void)
{	
  return (PSDCNT & 0x01);
}
//=======================================================
void GP_SendGPU0(u_long gpu_primitive)
{	
  gp0 = gpu_primitive;
}
//=======================================================
void ResetGraph(int mode)
{	
  gp1 = mode;
}
//=======================================================
long SetVideoMode(long mode)
{	
  u_long old_mode;

  old_mode = VIDEOMODE;

  VIDEOMODE = mode;

  if (mode > 1) {

    VIDEOMODE = MODE_NTSC;

    if (*(char *)0xbfc7ff52 == 'E')
      VIDEOMODE = MODE_PAL;

    }

  return (old_mode);
}
//=======================================================
long GetVideoMode(void)
{	
  return (VIDEOMODE);
}
//=======================================================
void GsDrawOT(u_long *pGP)
{	
  int i;
  u_long *OBJECT_last;

  if (OBJECT_number > 1) {

    OBJECT_last = OBJECT_end[OBJECT_number-1];

    object_quick_sort(0, OBJECT_number - 1);

    SWITCH_ADDRESS (OBJECT_first, OBJECT_start[0]);

    for (i = 1; i < OBJECT_number; i++) {
      SWITCH_ADDRESS (OBJECT_end[i-1], OBJECT_start[i]);
      }  

    SWITCH_ADDRESS (OBJECT_end[OBJECT_number-1], OBJECT_last);

    }

  gp1 = (0x05000000 | (PSDOFSY[PSDCNT & 0x01] << 10) | PSDOFSX[PSDCNT & 0x01]);

  *(G_GP) = 0x00ffffff;			// mark the end of the list

  d_icr = 0;

  d2_madr = (u_long) pGP;
  d2_bcr = 0;
  d2_chcr = d2_linked;

  PSDCNT++;

  OBJECT_number = 0;
}
//=======================================================
//=======================================================
void GsSortObject(GsOBJECT *object)
{
  OBJECT_start[OBJECT_number] = G_GP;

  if (OBJECT_number == 0)
    OBJECT_first = OBJECT_prev;

  // store the 'projected' z value
  OBJECT_z[OBJECT_number] = GsSortObject2(object);

  OBJECT_end[OBJECT_number] = OBJECT_prev;

  OBJECT_number++;

}
//=======================================================
void object_quick_sort(int s, int e)
{
  int h, g, l, r;
  u_long *os, *oe;

  if (s < e) {
    g = OBJECT_z[s + (e - s) / 2];
    l = s;
    r = e;

    while (l < r) {

      // This quick sort is descending, so that the largest Z coordinate is drawn first.
      // For an ascending sort, simply reverse the < > signs in the next 2 lines
      while ( (OBJECT_z[l] > g) && (l <= e) ) l++;
      while ( (OBJECT_z[r] < g) && (r >= s) ) r--;

      if (l <= r) {

        h = OBJECT_z[l];
        OBJECT_z[l] = OBJECT_z[r];
        OBJECT_z[r] = h;

	os = OBJECT_start[l];
	OBJECT_start[l] = OBJECT_start[r];
	OBJECT_start[r] = os;

	oe = OBJECT_end[l];
	OBJECT_end[l] = OBJECT_end[r];
	OBJECT_end[r] = oe;

        l++;
        r--;
        }

      }

    object_quick_sort(s, r);
    object_quick_sort(l, e);
    }

}
//=======================================================
void GsSortLine(GsLINE *line)
{	
  ADD_PACKET (0, GPU_COM (0x40, line->r, line->g, line->b));
  ADD_PACKET (1, H2L (line->x0, line->y0));
  ADD_PACKET (2, H2L (line->x1, line->y1));
  NEXT_PACKET (3);
}
//=======================================================
void GsSortGLine (GsGLINE *line)
{	
  ADD_PACKET (0, GPU_COM (0x50, line->r0, line->g0, line->b0));
  ADD_PACKET (1, H2L (line->x0, line->y0));
  ADD_PACKET (2, GPU_COM (0x00, line->r1, line->g1, line->b1));
  ADD_PACKET (3, H2L (line->x1, line->y1));
  NEXT_PACKET (4);
}
//=======================================================
void GsSortTile(GsTILE *tile)
{	
  int trans_on, trans_rate;

  trans_rate = tile->trans - 1;

  if (trans_rate < 0)
    trans_on = 0;
  else
    trans_on = 2;

  trans_rate = trans_rate << 5;

  ADD_PACKET (0, (0xe1000200 | trans_rate));
  ADD_PACKET (1, GPU_COM ((0x60 | trans_on), tile->r, tile->g, tile->b));
  ADD_PACKET (2, H2L (tile->x, tile->y));
  ADD_PACKET (3, H2L (tile->w, tile->h));
  NEXT_PACKET (4);
}
//=======================================================
void GsSortDot(GsDOT *dot)
{	
  ADD_PACKET (0, GPU_COM (0x68, dot->r, dot->g, dot->b));
  ADD_PACKET (1, H2L (dot->x, dot->y));
  NEXT_PACKET (2);
}
//=======================================================
void GsSortSprite(GsSPRITE *sprite)
{	
  //  int trans_on, trans_rate;
  //  trans_rate = sprite->trans - 1;
  //  if (trans_rate < 0)
  //trans_on = 0;
  //else
  //    trans_on = 1 << 25;
  //trans_rate = trans_rate << 5;
  //ADD_PACKET (0) = (0xe1000200 | trans_rate | sprite->tpage);

  ADD_PACKET (0, (0xe1000200 | sprite->tpage));

  if ((sprite->rotate % ONE) | (sprite->scalex % ONE) | (sprite->scalex % ONE))
    GsSortSprite2(sprite);
  else {
    ADD_PACKET (1, GPU_COM (0x64, sprite->r, sprite->g, sprite->b));
    ADD_PACKET (2, H2L (sprite->x, sprite->y));
    ADD_PACKET (3, HB2L (sprite->clutid, sprite->u, sprite->v));
    ADD_PACKET (4, H2L (sprite->w, sprite->h));
    NEXT_PACKET (5);
    }
}
//=======================================================
void GsSortSprite2(GsSPRITE *sprite)
{	
  GsTPOLY4 poly;
  long new_width, new_height;
  long offset_x, offset_y;
  long x0, y0, x1, y1, x2, y2, x3, y3;
  long cxa, cxb, cya, cyb;
  long sxa, sxb, sya, syb;
  short rot_cos, rot_sin;

  //int trans_on;

  //if ((sprite->trans - 1) < 0)
    //trans_on = 0;
  //else
    //trans_on = 1 << 25;

  // To rotate around the centre, first calculate the x/y coords of the
  // vertices as offsets from the centre.
  // Scaling is multiplied up by 4096 (ONE), so to get 'on-screen' scaled size we divide
  // again by 4096 ( >> 12).
  // To get offset from the centre, width and height are halved ( >> 1), meaning we use
  // a right-shift of 13 ( >> 13).
  // ie. the 'scaled' width calculation can be written as:
  // new_width = ((sprite->w * sprite->scalex) >> 12) / 2;

  new_width = ((sprite->w *sprite->scalex) >> 13);
  new_height = ((sprite->h *sprite->scaley) >> 13);
  if ((new_width > 255) | (new_height > 255))
    return;

  if ((new_width <= 0) | (new_height <= 0))
    return;

  offset_x = sprite->x + (sprite->w >> 1);
  offset_y = sprite->y + (sprite->h >> 1);

  rot_cos = cos(sprite->rotate);
  rot_sin = sin(sprite->rotate);

  x0 = x2 = 0 - new_width;
  y0 = y1 = 0 - new_height;
  x1 = x3 = 0 + new_width;
  y2 = y3 = 0 + new_height;

  cxa = (x0 * rot_cos) >> 12;
  sxa = (x0 * rot_sin) >> 12;
  cya = (y0 * rot_cos) >> 12;
  sya = (y0 * rot_sin) >> 12;

  cxb = (x1 * rot_cos) >> 12;
  sxb = (x1 * rot_sin) >> 12;
  cyb = (y2 * rot_cos) >> 12;
  syb = (y2 * rot_sin) >> 12;

  poly.x0 = cxa - sya + offset_x;
  poly.y0 = cya + sxa + offset_y;
  poly.x1 = cxb - sya + offset_x;
  poly.y1 = cya + sxb + offset_y;
  poly.x2 = cxa - syb + offset_x;
  poly.y2 = cyb + sxa + offset_y;
  poly.x3 = cxb - syb + offset_x;
  poly.y3 = cyb + sxb + offset_y;

  poly.u1 = poly.u3 = sprite->u + sprite->w - 1;
  poly.v2 = poly.v3 = sprite->v + sprite->h - 1;

  if (poly.u1 > 255)
    poly.u1 = poly.u3 = 255;

  if (poly.v2 > 255)
    poly.v2 = poly.v3 = 255;

  ADD_PACKET (1, GPU_COM (0x2c, sprite->r, sprite->g, sprite->b));
  ADD_PACKET (2, H2L (poly.x0, poly.y0));
  ADD_PACKET (3, HB2L (sprite->clutid, sprite->u, sprite->v));
  ADD_PACKET (4, H2L (poly.x1, poly.y1));
  ADD_PACKET (5, HB2L (sprite->tpage, poly.u1, sprite->v));
  ADD_PACKET (6, H2L (poly.x2, poly.y2));
  ADD_PACKET (7, HB2L (0, sprite->u, poly.v2));
  ADD_PACKET (8, H2L (poly.x3, poly.y3));
  ADD_PACKET (9, HB2L (0, poly.u3, poly.v3));
  NEXT_PACKET (10);

}
//=======================================================
void GsSortPoly3 (GsPOLY3 *poly)
{	
  ADD_PACKET (0, 0xe1000200);
  ADD_PACKET (1, GPU_COM (0x20, poly->r, poly->g, poly->b));
  ADD_PACKET (2, H2L (poly->x0, poly->y0));
  ADD_PACKET (3, H2L (poly->x1, poly->y1));
  ADD_PACKET (4, H2L (poly->x2, poly->y2));
  NEXT_PACKET (5);
}
//=======================================================
void GsSortTPoly3(GsTPOLY3 *poly)
{	
  ADD_PACKET (0, 0xe1000200);
  ADD_PACKET (1, GPU_COM (0x24, poly->r, poly->g, poly->b));
  ADD_PACKET (2, H2L (poly->x0, poly->y0));
  ADD_PACKET (3, HB2L (poly->clutid, poly->u0, poly->v0));
  ADD_PACKET (4, H2L (poly->x1, poly->y1));
  ADD_PACKET (5, HB2L (poly->tpage, poly->u1, poly->v1));
  ADD_PACKET (6, H2L (poly->x2, poly->y2));
  ADD_PACKET (7, HB2L (0, poly->u2, poly->v2));
  NEXT_PACKET (8);
}
//=======================================================
void GsSortPoly4(GsPOLY4 *poly)
{	
  ADD_PACKET (0, GPU_COM (0x28, poly->r, poly->g, poly->b));
  ADD_PACKET (1, H2L (poly->x0, poly->y0));
  ADD_PACKET (2, H2L (poly->x1, poly->y1));
  ADD_PACKET (3, H2L (poly->x2, poly->y2));
  ADD_PACKET (4, H2L (poly->x3, poly->y3));
  NEXT_PACKET (5);
}
//=======================================================
void GsSortTPoly4(GsTPOLY4 *poly)
{	
  ADD_PACKET (0, 0xe1000200);
  ADD_PACKET (1, GPU_COM (0x2c, poly->r, poly->g, poly->b));
  ADD_PACKET (2, H2L (poly->x0, poly->y0));
  ADD_PACKET (3, HB2L (poly->clutid, poly->u0, poly->v0));
  ADD_PACKET (4, H2L (poly->x1, poly->y1));
  ADD_PACKET (5, HB2L (poly->tpage, poly->u1, poly->v1));
  ADD_PACKET (6, H2L (poly->x2, poly->y2));
  ADD_PACKET (7, HB2L (0, poly->u2, poly->v2));
  ADD_PACKET (8, H2L (poly->x3, poly->y3));
  ADD_PACKET (9, HB2L (0, poly->u3, poly->v3));
  NEXT_PACKET (10);
}
//=======================================================
void GsSortGTPoly4(GsGTPOLY4 *poly)
{	
  ADD_PACKET (0, 0xe1000200);
  ADD_PACKET (1, GPU_COM (0x3c, poly->r0, poly->g0, poly->b0));
  ADD_PACKET (2, H2L (poly->x0, poly->y0));
  ADD_PACKET (3, HB2L (poly->clutid, poly->u0, poly->v0));
  ADD_PACKET (4, GPU_COM (0x00, poly->r1, poly->g1, poly->b1));
  ADD_PACKET (5, H2L (poly->x1, poly->y1));
  ADD_PACKET (6, HB2L (poly->tpage, poly->u1, poly->v1));
  ADD_PACKET (7, GPU_COM (0x00, poly->r2, poly->g2, poly->b2));
  ADD_PACKET (8, H2L (poly->x2, poly->y2));
  ADD_PACKET (9, HB2L (0, poly->u2, poly->v2));
  ADD_PACKET (10, GPU_COM (0x00, poly->r3, poly->g3, poly->b3));
  ADD_PACKET (11, H2L (poly->x3, poly->y3));
  ADD_PACKET (12, HB2L (0, poly->u3, poly->v3));
  NEXT_PACKET (13);
}
//=======================================================
void GsSortGPoly3(GsGPOLY3 *poly)
{	
  ADD_PACKET (0, 0xe1000200);
  ADD_PACKET (1, GPU_COM (0x30, poly->r0, poly->g0, poly->b0));
  ADD_PACKET (2, H2L (poly->x0, poly->y0));
  ADD_PACKET (3, GPU_COM (0x00, poly->r1, poly->g1, poly->b1));
  ADD_PACKET (4, H2L (poly->x1, poly->y1));
  ADD_PACKET (5, GPU_COM (0x00, poly->r2, poly->g2, poly->b2));
  ADD_PACKET (6, H2L (poly->x2, poly->y2));
  NEXT_PACKET (7);
}
//=======================================================
void GsSortGTPoly3(GsGTPOLY3 *poly)
{	
  ADD_PACKET (0, (0xe1000200));
  ADD_PACKET (1, GPU_COM (0x34, poly->r0, poly->g0, poly->b0));
  ADD_PACKET (2, H2L (poly->x0, poly->y0));
  ADD_PACKET (3, HB2L (poly->clutid, poly->u0, poly->v0));
  ADD_PACKET (4, GPU_COM (0x00, poly->r1, poly->g1, poly->b1));
  ADD_PACKET (5, H2L (poly->x1, poly->y1));
  ADD_PACKET (6, HB2L (poly->tpage, poly->u1, poly->v1));
  ADD_PACKET (7, GPU_COM (0x00, poly->r2, poly->g2, poly->b2));
  ADD_PACKET (8, H2L (poly->x2, poly->y2));
  ADD_PACKET (9, HB2L (0, poly->u2, poly->v2));
  NEXT_PACKET (10);
}
//=======================================================
void GsSortGPoly4(GsGPOLY4 *poly)
{	
  ADD_PACKET (0, (0xe1000200));
  ADD_PACKET (1, GPU_COM (0x38, poly->r0, poly->g0, poly->b0));
  ADD_PACKET (2, H2L (poly->x0, poly->y0));
  ADD_PACKET (3, GPU_COM (0x00, poly->r1, poly->g1, poly->b1));
  ADD_PACKET (4, H2L (poly->x1, poly->y1));
  ADD_PACKET (5, GPU_COM (0x00, poly->r2, poly->g2, poly->b2));
  ADD_PACKET (6, H2L (poly->x2, poly->y2));
  ADD_PACKET (7, GPU_COM (0x00, poly->r3, poly->g3, poly->b3));
  ADD_PACKET (8, H2L (poly->x3, poly->y3));
  NEXT_PACKET (9);
}
//=======================================================
void GsSetWorkBase(u_char *gp_addr)
{	
  G_GP = (u_long *) gp_addr;
}
//=======================================================
void GsSortClear(int r, int g, int b)
{	
  r &= 0xff;
  g &= 0xff;
  b &= 0xff;
  ADD_PACKET (0, GPU_COM (0x60, r, g, b));
  ADD_PACKET (1, 0x00000000);
  ADD_PACKET (2, H2L (GsDISPENV.disp.w, GsDISPENV.disp.h));
  NEXT_PACKET (3);
}
//=======================================================
void GsSwapDispBuff(void)
{	
  u_long x, y, data;

  x = PSDOFSX[PSDCNT & 0x01];
  y = PSDOFSY[PSDCNT & 0x01];

  data = ((y + GsDISPENV.disp.h - 1) << 10) | (x + GsDISPENV.disp.w);

  ADD_PACKET (0, (0xe5000000 | (y << 11) | x));
  ADD_PACKET (1, (0xe3000000 | (y << 10) | x));
  ADD_PACKET (2, (0xe4000000 | data));
  NEXT_PACKET (3);
}
//=======================================================
void GsGetTimInfo(TIMHEADER *tim, GsIMAGE *img)
{	
  TIMDATA *clut = NULL;
  TIMDATA *pixel = NULL;
  u_long *pp = NULL;

  // get the color mode
  img->pmode = (tim->flag & 0x03);

  if ((tim->flag & 0x08) == 0) {		// there is no clut
    pixel = (TIMDATA *)++tim;
    img->clutid = 0;
    }
    else {					// there is a clut
      clut = (TIMDATA *)++tim;
      img->clutid = ((clut->vrampos >> 10) | (clut->vrampos & 0xffff) >> 4);
      pp = (u_long *) clut + (clut->datalen >> 2);
      pixel = (TIMDATA *) pp;
      }

  // x/y address of pixel data
  img->px = (pixel->vrampos & 0xffff);
  img->py = (pixel->vrampos >> 16);

  // tpage number
  img->tpage = ((img->py & 0x100) >> 4) | ((img->px & 0x3ff) >> 6) | ((img->py & 0x200) << 2);
  img->tpage = ((img->pmode << 7) | img->tpage);

  img->px %= 64;				  // x/y offset within tpage
  if (img->py > 255)
    img->py -= 256;

  img->pw = (pixel->size & 0xffff) << (2 - img->pmode);		  // width/height of pixel data
  img->ph = (pixel->size >> 16);
}
//=======================================================
void LoadImage(RECT *rect, u_long *p)
{	
  u_long vrampos = (rect->y << 16 | rect->x);
  u_long size = (rect->h << 16 | rect->w);
  u_long size_in_words = (rect->w *rect->h);
  u_long i;

  wait_gpu;

  gp1 = 0x01000000;			// reset the command queue
  GP_SendGPU0(0xa0000000);		// send 'send image' primitive
  GP_SendGPU0(vrampos);			// send x and y coordinates
  GP_SendGPU0(size);			// send width and height

  for (i = 0; i < size_in_words; i++)
    GP_SendGPU0(*(p + i));

  gp1 = 0x01000000;			// reset the command queue again
}
//=======================================================
long LoadTim(TIMHEADER *tim)
{	
  TIMDATA *clut = NULL;
  TIMDATA *pixel = NULL;
  u_long *pp = NULL;

  // is it a valid TIM file?
  if (tim->id != 0x10)
    return (- 1);

  // is there a CLUT?
  if ((tim->flag & 0x08) == 0) {
    pixel = (TIMDATA *)++tim;
    Mem2Vram(pixel, pixel->datalen);
    }
  else {
    clut = (TIMDATA *)++tim;
    Mem2Vram(clut, clut->datalen);
    pp = (u_long *) clut + (clut->datalen >> 2);
    pixel = (TIMDATA *) pp;
    Mem2Vram(pixel, pixel->datalen);
    }
  return (0);
}
//=======================================================
void Mem2Vram(TIMDATA *pTIM, u_long size_in_bytes)
{	
  u_long i;
  u_long *pp = (u_long *) pTIM;
  u_long size_in_words = (size_in_bytes >> 2) - 3;

  wait_gpu;

  gp1 = 0x01000000;			// reset the command queue
  GP_SendGPU0(0xa0000000);		// send 'send image' primitive
  GP_SendGPU0(*(pp + 1));		// send x and y coordinates
  GP_SendGPU0(*(pp + 2));		// send width and height

  for (i = 0; i < size_in_words; i++)
    GP_SendGPU0(*(pp + 3 + i));

  gp1 = 0x01000000;			// reset the command queue again
}
//=======================================================
void FntColor(u_char num, u_char r, u_char g, u_char b)
{	
  FONTCOLOR[num] = (0x74000000 | (b << 16) | (g << 8) | r);
}
//=======================================================
void FntLoad(u_char num, TIMHEADER *new_font)
{	
  if (new_font == NULL)
    new_font = (TIMHEADER *) specfont;

  GsGetTimInfo((TIMHEADER *) new_font, &FONT[num]);
  LoadTim((TIMHEADER *) new_font);

  // set RGB to mid values
  FONTCOLOR[num] = 0x74808080;

  // set width
  FONT_WIDTH[num] = 0;

  // set height
  FONT_HEIGHT[num] = 0;
}
//=======================================================
void FreeFntLoad(u_char num, u_char width, u_char height, TIMHEADER *new_font)
{	
  if (new_font == NULL)
    new_font = (TIMHEADER *) specfont;

  GsGetTimInfo((TIMHEADER *) new_font, &FONT[num]);
  LoadTim((TIMHEADER *) new_font);

  // set RGB to mid values
  FONTCOLOR[num] = 0x74808080;

  // set width
  FONT_WIDTH[num] = width;

  // set height
  FONT_HEIGHT[num] = height;
}
//=======================================================
void FntPrint(u_char num, int x, int y, char *fmt, ...)
{	
  int strn = 0;
  int count = 0;
  int c = 0;
  int w = 0;
  int padc = 0; 
  int i = 0; 
  int n = 0; 
  char *s = NULL;
  char *out = NULL;
  va_list argptr;

  va_start(argptr, fmt);

  out = STRINGBUFF + strn;

  while ((c = *fmt++) != 0) {

    if (c != '%') {
      out[count++] = c;
      continue;
      }

    if (*fmt == '-')
      fmt++;

    padc = (*fmt == '0') ? '0' : ' ';

    for (w = 0; *fmt >= '0' && *fmt <= '9'; fmt++)
	     w = w *10 + *fmt - '0';

    if (! *fmt)
      break;

    switch (*fmt++) {
      case 'x':	
		n = argptr[0];
		count += PutNum(out + count, n, w, padc, 16);
      		break;
      case 'd':	
		n = argptr[0];
      		count += PutNum(out + count, n, w, padc, 10);
      		break;
      case 'c':	
		c = argptr[0];
      		out[count++] = c;
      		break;
      case 's':	
		out = argptr[0]; 
	       	  
		/*	out[count++] = *s;
      		*/	
/*			while (w-- > 0)
			{ out[count++] = ' ';}*/
		

		break;
      }

    }


  strn += count;

  va_end(argptr);

  /* set the texture page for the font */
  ADD_PACKET (0, (0xe1000200 | FONT[num].tpage));
  NEXT_PACKET (1);

  for (n = 0; n < count; n++) {
    PrintChar(num, x, y, *out);
    out++;
    x += 8;
    }

}
//=======================================================
int PutNum(char *out, int n, int w, int padc, int radix)
{	
  int i, minus, cnt;
  char buf[128];

  static char digit[16] = "0123456789abcdef";
  minus = (n < 0);
  if (minus)
    n = - n;

  for (i = 0; n; i++) {
    buf[i] = digit[(n % radix)];
    n /= radix;
    }
  if (!i)
    buf[i++] = '0';

  if (minus) {
    if (padc == ' ') {
      buf[i++] = '-';
      }
    else
      w--;

    }
  while (i < w)
    buf[i++] = padc;

  if (minus && padc != ' ')
    buf[i++] = '-';

  cnt = 0;

  while (--i >= 0) {
    *out++ = buf[i];
    cnt++;
    }

  return (cnt);
}
//=======================================================
void PrintChar(u_char num, int x, int y, u_char ascii)
{	
  u_char u, v;

  if (FONT_WIDTH[num] + FONT_WIDTH[num] == 0) {

    /* if the font is only 16 pixels high, it cannot include lower case, so convert to upper */
    if ((FONT[num].ph == 16) && (ascii > 0x60))
      ascii -= 0x20;

    /* work out the character offset in the TIM image */
    u = (ascii & 0x1f) << 3;
    v = (((ascii & 0xe0) >> 5) - 1) << 3;

    /* work out the offset in the texture page */
    u += FONT[num].px;
    v += FONT[num].py;

    /* add the character to the Ordering Table */
    ADD_PACKET (0, FONTCOLOR[num]);
    ADD_PACKET (1, H2L (x, y));
    ADD_PACKET (2, HB2L (FONT[num].clutid, u, v));
    NEXT_PACKET (3);
    }
}
//=======================================================
